# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto

PACKAGE = "a.test.package.with.manifest"
__protobuf__ = proto.module(package=PACKAGE, manifest={"ThisFoo", "ThisBar"},)


class ThisFoo(proto.Message):
    foo = proto.Field(proto.INT32, number=1)


class ThisBar(proto.Message):
    bar = proto.Field(proto.INT32, number=2)


def test_manifest_causes_exclusion_of_classname_salt():
    assert (
        ThisFoo.pb(ThisFoo()).DESCRIPTOR.file.name
        == "test_message_filename_with_manifest.proto"
    )
    assert (
        ThisBar.pb(ThisBar()).DESCRIPTOR.file.name
        == "test_message_filename_with_manifest.proto"
    )



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

from google.protobuf import descriptor_pb2

from proto import _file_info
from proto import _package_info
from proto.marshal.rules.enums import EnumRule


class ProtoEnumMeta(enum.EnumMeta):
    """A metaclass for building and registering protobuf enums."""

    def __new__(mcls, name, bases, attrs):
        # Do not do any special behavior for `proto.Enum` itself.
        if bases[0] == enum.IntEnum:
            return super().__new__(mcls, name, bases, attrs)

        # Get the essential information about the proto package, and where
        # this component belongs within the file.
        package, marshal = _package_info.compile(name, attrs)

        # Determine the local path of this proto component within the file.
        local_path = tuple(attrs.get("__qualname__", name).split("."))

        # Sanity check: We get the wrong full name if a class is declared
        # inside a function local scope; correct this.
        if "<locals>" in local_path:
            ix = local_path.index("<locals>")
            local_path = local_path[: ix - 1] + local_path[ix + 1 :]

        # Determine the full name in protocol buffers.
        full_name = ".".join((package,) + local_path).lstrip(".")
        filename = _file_info._FileInfo.proto_file_name(
            attrs.get("__module__", name.lower())
        )

        # Retrieve any enum options.
        # We expect something that looks like an EnumOptions message,
        # either an actual instance or a dict-like representation.
        pb_options = "_pb_options"
        opts = attrs.pop(pb_options, {})
        # This is the only portable way to remove the _pb_options name
        # from the enum attrs.
        # In 3.7 onwards, we can define an _ignore_ attribute and do some
        # mucking around with that.
        if pb_options in attrs._member_names:
            idx = attrs._member_names.index(pb_options)
            attrs._member_names.pop(idx)

        # Make the descriptor.
        enum_desc = descriptor_pb2.EnumDescriptorProto(
            name=name,
            # Note: the superclass ctor removes the variants, so get them now.
            # Note: proto3 requires that the first variant value be zero.
            value=sorted(
                (
                    descriptor_pb2.EnumValueDescriptorProto(name=name, number=number)
                    # Minor hack to get all the enum variants out.
                    for name, number in attrs.items()
                    if isinstance(number, int)
                ),
                key=lambda v: v.number,
            ),
            options=opts,
        )

        file_info = _file_info._FileInfo.maybe_add_descriptor(filename, package)
        if len(local_path) == 1:
            file_info.descriptor.enum_type.add().MergeFrom(enum_desc)
        else:
            file_info.nested_enum[local_path] = enum_desc

        # Run the superclass constructor.
        cls = super().__new__(mcls, name, bases, attrs)

        # We can't just add a "_meta" element to attrs because the Enum
        # machinery doesn't know what to do with a non-int value.
        # The pb is set later, in generate_file_pb
        cls._meta = _EnumInfo(full_name=full_name, pb=None)

        file_info.enums[full_name] = cls

        # Register the enum with the marshal.
        marshal.register(cls, EnumRule(cls))

        # Generate the descriptor for the file if it is ready.
        if file_info.ready(new_class=cls):
            file_info.generate_file_pb(new_class=cls, fallback_salt=full_name)

        # Done; return the class.
        return cls


class Enum(enum.IntEnum, metaclass=ProtoEnumMeta):
    """A enum object that also builds a protobuf enum descriptor."""

    pass


class _EnumInfo:
    def __init__(self, *, full_name: str, pb):
        self.full_name = full_name
        self.pb = pb



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import proto


def test_bytes_init():
    class Foo(proto.Message):
        bar = proto.Field(proto.BYTES, number=1)
        baz = proto.Field(proto.BYTES, number=2)

    foo = Foo(bar=b"spam")
    assert foo.bar == b"spam"
    assert foo.baz == b""
    assert not foo.baz
    assert Foo.pb(foo).bar == b"spam"
    assert Foo.pb(foo).baz == b""


def test_bytes_rmw():
    class Foo(proto.Message):
        spam = proto.Field(proto.BYTES, number=1)
        eggs = proto.Field(proto.BYTES, number=2)

    foo = Foo(spam=b"bar")
    foo.eggs = b"baz"
    assert foo.spam == b"bar"
    assert foo.eggs == b"baz"
    assert Foo.pb(foo).spam == b"bar"
    assert Foo.pb(foo).eggs == b"baz"
    foo.spam = b"bacon"
    assert foo.spam == b"bacon"
    assert foo.eggs == b"baz"
    assert Foo.pb(foo).spam == b"bacon"
    assert Foo.pb(foo).eggs == b"baz"


def test_bytes_del():
    class Foo(proto.Message):
        bar = proto.Field(proto.BYTES, number=1)

    foo = Foo(bar=b"spam")
    assert foo.bar == b"spam"
    del foo.bar
    assert foo.bar == b""
    assert not foo.bar


def test_bytes_string_distinct():
    class Foo(proto.Message):
        bar = proto.Field(proto.STRING, number=1)
        baz = proto.Field(proto.BYTES, number=2)

    foo = Foo()
    assert foo.bar != foo.baz

    # Since protobuf was written against Python 2, it accepts bytes objects
    # for strings (but not vice versa).
    foo.bar = b"anything"
    assert foo.bar == "anything"
    with pytest.raises(TypeError):
        foo.baz = "anything"



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


def test_composite_init():
    class Foo(proto.Message):
        bar = proto.Field(proto.STRING, number=1)
        baz = proto.Field(proto.INT64, number=2)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)
        eggs = proto.Field(proto.BOOL, number=2)

    spam = Spam(foo=Foo(bar="str", baz=42))
    assert spam.foo.bar == "str"
    assert spam.foo.baz == 42
    assert spam.eggs is False


def test_composite_inner_rmw():
    class Foo(proto.Message):
        bar = proto.Field(proto.STRING, number=1)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)

    spam = Spam(foo=Foo(bar="str"))
    spam.foo.bar = "other str"
    assert spam.foo.bar == "other str"
    assert Spam.pb(spam).foo.bar == "other str"


def test_composite_empty_inner_rmw():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)

    spam = Spam()
    spam.foo.bar = 42
    assert spam.foo.bar == 42


def test_composite_outer_rmw():
    class Foo(proto.Message):
        bar = proto.Field(proto.FLOAT, number=1)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)

    spam = Spam(foo=Foo(bar=3.14159))
    spam.foo = Foo(bar=2.71828)
    assert abs(spam.foo.bar - 2.71828) < 1e-7


def test_composite_dict_write():
    class Foo(proto.Message):
        bar = proto.Field(proto.FLOAT, number=1)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)

    spam = Spam()
    spam.foo = {"bar": 2.71828}
    assert abs(spam.foo.bar - 2.71828) < 1e-7


def test_composite_del():
    class Foo(proto.Message):
        bar = proto.Field(proto.STRING, number=1)

    class Spam(proto.Message):
        foo = proto.Field(proto.MESSAGE, number=1, message=Foo)

    spam = Spam(foo=Foo(bar="str"))
    del spam.foo
    assert not spam.foo
    assert isinstance(spam.foo, Foo)
    assert spam.foo.bar == ""



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from google.protobuf import struct_pb2

import proto


def test_value_primitives_read():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    assert Foo(value=3).value == 3.0
    assert Foo(value="3").value == "3"
    assert Foo(value=None).value is None
    assert Foo(value=False).value is False
    assert Foo(value=True).value is True


def test_value_absent():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    assert Foo().value is None


def test_value_primitives_rmw():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo()
    foo.value = 3
    assert isinstance(foo.value, float)
    assert abs(Foo.pb(foo).value.number_value - 3.0) < 1e-7
    foo.value = False
    assert not foo.value
    assert foo.value is False
    assert Foo.pb(foo).value.WhichOneof("kind") == "bool_value"
    foo.value = None
    assert not foo.value
    assert foo.value is None
    assert Foo.pb(foo).value.WhichOneof("kind") == "null_value"


def test_value_write_pb():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value=struct_pb2.Value(string_value="stringy"))
    assert foo.value == "stringy"


def test_value_lists_read():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value=["3", None, "foo", True])
    assert foo.value == ["3", None, "foo", True]


def test_value_lists_null():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.ListValue, number=1)

    foo = Foo()
    assert foo.value is None


def test_value_struct_null():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Struct, number=1)

    foo = Foo()
    assert foo.value is None


def test_value_lists_rmw():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value=["3", None, "foo", True])
    foo.value.append("bar")
    foo.value.pop(1)
    assert foo.value == ["3", "foo", True, "bar"]


def test_value_lists_nested():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value=[[True, False], [True, False]])
    foo.value.append([False, True])
    assert foo.value == [[True, False], [True, False], [False, True]]


def test_value_lists_struct():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value=[{"foo": "bar", "spam": "eggs"}])
    assert foo.value == [{"foo": "bar", "spam": "eggs"}]


def test_value_lists_detachment():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value=["foo", "bar"])
    detached_list = foo.value
    detached_list.append("baz")
    assert foo.value == ["foo", "bar", "baz"]


def test_value_structs_read():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value={"foo": True, "bar": False})
    assert foo.value == {"foo": True, "bar": False}


def test_value_structs_rmw():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value={"foo": True, "bar": False})
    foo.value["baz"] = "a string"
    assert foo.value == {"foo": True, "bar": False, "baz": "a string"}


def test_value_structs_nested():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo(value={"foo": True, "bar": {"spam": "eggs"}})
    assert foo.value == {"foo": True, "bar": {"spam": "eggs"}}
    sv = Foo.pb(foo).value.struct_value
    assert sv["foo"] is True
    assert sv["bar"].fields["spam"].string_value == "eggs"


def test_value_invalid_value():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    with pytest.raises(ValueError):
        Foo(value=object())


def test_value_unset():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Value, number=1)

    foo = Foo()
    assert "value" not in foo


def test_list_value_read():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.ListValue, number=1)

    foo = Foo(value=["foo", "bar", True, {"spam": "eggs"}])
    assert foo.value == ["foo", "bar", True, {"spam": "eggs"}]


def test_list_value_pb():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.ListValue, number=1)

    foo = Foo(
        value=struct_pb2.ListValue(
            values=[
                struct_pb2.Value(string_value="foo"),
                struct_pb2.Value(string_value="bar"),
                struct_pb2.Value(bool_value=True),
            ]
        )
    )
    assert foo.value == ["foo", "bar", True]


def test_list_value_reassignment():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.ListValue, number=1)

    foo = Foo(value=["foo", "bar"])
    detached = foo.value
    detached.append(True)
    foo.value = detached
    assert foo.value == ["foo", "bar", True]


def test_list_value_invalid():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.ListValue, number=1)

    with pytest.raises(TypeError):
        Foo(value=3)


def test_struct_read():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Struct, number=1)

    foo = Foo(value={"foo": "bar", "bacon": True})
    assert foo.value == {"foo": "bar", "bacon": True}


def test_struct_pb():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Struct, number=1)

    foo = Foo(
        value=struct_pb2.Struct(
            fields={
                "foo": struct_pb2.Value(string_value="bar"),
                "bacon": struct_pb2.Value(bool_value=True),
            }
        )
    )
    assert foo.value == {"foo": "bar", "bacon": True}


def test_struct_reassignment():
    class Foo(proto.Message):
        value = proto.Field(struct_pb2.Struct, number=1)

    foo = Foo(value={"foo": "bar"})
    detached = foo.value
    detached["bacon"] = True
    foo.value = detached
    assert foo.value == {"foo": "bar", "bacon": True}



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


PACKAGE = "a.test.package.with.and.without.manifest"
__protobuf__ = proto.module(package=PACKAGE, manifest={"This", "That"},)


class This(proto.Message):
    this = proto.Field(proto.INT32, number=1)


class That(proto.Message):
    that = proto.Field(proto.INT32, number=1)


class NotInManifest(proto.Message):
    them = proto.Field(proto.INT32, number=1)


def test_manifest_causes_exclusion_of_classname_salt():

    assert (
        This.pb(This()).DESCRIPTOR.file.name
        == "test_message_filename_with_and_without_manifest.proto"
    )
    assert (
        That.pb(That()).DESCRIPTOR.file.name
        == "test_message_filename_with_and_without_manifest.proto"
    )

    assert (
        NotInManifest.pb(NotInManifest()).DESCRIPTOR.file.name
        == "test_message_filename_with_and_without_manifest_"
        + PACKAGE
        + ".notinmanifest.proto"
    )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import proto


def test_int_init():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)
        baz = proto.Field(proto.INT32, number=2)

    foo = Foo(bar=42)
    assert foo.bar == 42
    assert foo.baz == 0
    assert not foo.baz
    assert Foo.pb(foo).bar == 42
    assert Foo.pb(foo).baz == 0


def test_int_rmw():
    class Foo(proto.Message):
        spam = proto.Field(proto.INT32, number=1)
        eggs = proto.Field(proto.INT32, number=2)

    foo = Foo(spam=42)
    foo.eggs = 76  # trombones led the big parade...
    assert foo.spam == 42
    assert foo.eggs == 76
    assert Foo.pb(foo).spam == 42
    assert Foo.pb(foo).eggs == 76
    foo.spam = 144
    assert foo.spam == 144
    assert foo.eggs == 76
    assert Foo.pb(foo).spam == 144
    assert Foo.pb(foo).eggs == 76


def test_int_del():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    foo = Foo(bar=42)
    assert foo.bar == 42
    del foo.bar
    assert foo.bar == 0
    assert not foo.bar


def test_int_size():
    class Foo(proto.Message):
        small = proto.Field(proto.INT32, number=1)
        big = proto.Field(proto.INT64, number=2)

    foo = Foo()
    foo.big = 2 ** 40
    assert foo.big == 2 ** 40
    with pytest.raises(ValueError):
        foo.small = 2 ** 40
    with pytest.raises(ValueError):
        Foo(small=2 ** 40)


def test_int_unsigned():
    class Foo(proto.Message):
        signed = proto.Field(proto.INT32, number=1)
        unsigned = proto.Field(proto.UINT32, number=2)

    foo = Foo()
    foo.signed = -10
    assert foo.signed == -10
    with pytest.raises(ValueError):
        foo.unsigned = -10
    with pytest.raises(ValueError):
        Foo(unsigned=-10)


def test_field_descriptor_idempotent():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    bar_field = Foo.meta.fields["bar"]
    assert bar_field.descriptor is bar_field.descriptor



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.protobuf import wrappers_pb2


class WrapperRule:
    """A marshal for converting the protobuf wrapper classes to Python.

    This class converts between ``google.protobuf.BoolValue``,
    ``google.protobuf.StringValue``, and their siblings to the appropriate
    Python equivalents.

    These are effectively similar to the protobuf primitives except
    that None becomes a possible value.
    """

    def to_python(self, value, *, absent: bool = None):
        if isinstance(value, self._proto_type):
            if absent:
                return None
            return value.value
        return value

    def to_proto(self, value):
        if isinstance(value, self._python_type):
            return self._proto_type(value=value)
        return value


class DoubleValueRule(WrapperRule):
    _proto_type = wrappers_pb2.DoubleValue
    _python_type = float


class FloatValueRule(WrapperRule):
    _proto_type = wrappers_pb2.FloatValue
    _python_type = float


class Int64ValueRule(WrapperRule):
    _proto_type = wrappers_pb2.Int64Value
    _python_type = int


class UInt64ValueRule(WrapperRule):
    _proto_type = wrappers_pb2.UInt64Value
    _python_type = int


class Int32ValueRule(WrapperRule):
    _proto_type = wrappers_pb2.Int32Value
    _python_type = int


class UInt32ValueRule(WrapperRule):
    _proto_type = wrappers_pb2.UInt32Value
    _python_type = int


class BoolValueRule(WrapperRule):
    _proto_type = wrappers_pb2.BoolValue
    _python_type = bool


class StringValueRule(WrapperRule):
    _proto_type = wrappers_pb2.StringValue
    _python_type = str


class BytesValueRule(WrapperRule):
    _proto_type = wrappers_pb2.BytesValue
    _python_type = bytes
